@@ -183,6 +183,13 @@ static struct builtin bintab[] = {
     BUILTIN("sched", 0, bin_sched, 0, -1, 0, NULL, NULL),
 };
 
+/**/
+int
+setup_sched(Module m)
+{
+    return 0;
+}
+
 /**/
 int
 boot_sched(Module m)
@@ -211,4 +218,11 @@ cleanup_sched(Module m)
     return 0;
 }
 
+/**/
+int
+finish_sched(Module m)
+{
+    return 0;
+}
+
 #endif
