@@ -42,7 +42,7 @@ int lastbase;
  
 static char *ptr;
 
-static long yyval;
+static zlong yyval;
 static LV yylval;
 
 static int mlevel = 0;
@@ -357,9 +357,11 @@ zzlex(void)
 	    }
 	    if (*ptr == '#') {
 		if (*++ptr == '\\') {
+		    int v;
+
 		    ptr++;
-		    yyval = *ptr == Meta ? *++ptr ^ 32 : *ptr;
-		    ptr++;
+		    ptr = getkeystring(ptr, NULL, 6, &v);
+		    yyval = v;
 		    unary = 0;
 		    return NUM;
 		}
@@ -409,14 +411,14 @@ static int sp = -1;			/* stack pointer */
 
 struct mathvalue {
     LV lval;
-    long val;
+    zlong val;
 };
 
 static struct mathvalue *stack;
 
 /**/
 static void
-push(long val, LV lval)
+push(zlong val, LV lval)
 {
     if (sp == STACKSZ - 1)
 	zerr("stack overflow", NULL, 0);
@@ -428,7 +430,7 @@ push(long val, LV lval)
 
 
 /**/
-static long
+static zlong
 getcvar(LV s)
 {
     char *t;
@@ -440,8 +442,8 @@ getcvar(LV s)
 
 
 /**/
-static long
-setvar(LV s, long v)
+static zlong
+setvar(LV s, zlong v)
 {
     if (s == -1 || s >= lvc) {
 	zerr("lvalue required", NULL, 0);
@@ -456,7 +458,7 @@ setvar(LV s, long v)
 
 /**/
 static int
-notzero(long a)
+notzero(zlong a)
 {
     if (a == 0) {
 	zerr("division by zero", NULL, 0);
@@ -496,7 +498,7 @@ notzero(long a)
 void
 op(int what)
 {
-    long a, b, c;
+    zlong a, b, c;
     LV lv;
 
     if (sp < 0) {
@@ -569,39 +571,39 @@ op(int what)
 	break;
     case LES:
 	pop2();
-	pushv((long)(a < b));
+	pushv((zlong)(a < b));
 	break;
     case LEQ:
 	pop2();
-	pushv((long)(a <= b));
+	pushv((zlong)(a <= b));
 	break;
     case GRE:
 	pop2();
-	pushv((long)(a > b));
+	pushv((zlong)(a > b));
 	break;
     case GEQ:
 	pop2();
-	pushv((long)(a >= b));
+	pushv((zlong)(a >= b));
 	break;
     case DEQ:
 	pop2();
-	pushv((long)(a == b));
+	pushv((zlong)(a == b));
 	break;
     case NEQ:
 	pop2();
-	pushv((long)(a != b));
+	pushv((zlong)(a != b));
 	break;
     case DAND:
 	pop2();
-	pushv((long)(a && b));
+	pushv((zlong)(a && b));
 	break;
     case DOR:
 	pop2();
-	pushv((long)(a || b));
+	pushv((zlong)(a || b));
 	break;
     case DXOR:
 	pop2();
-	pushv((long)((a && !b) || (!a && b)));
+	pushv((zlong)((a && !b) || (!a && b)));
 	break;
     case QUEST:
 	pop3();
@@ -658,15 +660,15 @@ op(int what)
 	break;
     case DANDEQ:
 	pop2lv();
-	set((long)(a && b));
+	set((zlong)(a && b));
 	break;
     case DOREQ:
 	pop2lv();
-	set((long)(a || b));
+	set((zlong)(a || b));
 	break;
     case DXOREQ:
 	pop2lv();
-	set((long)((a && !b) || (!a && b)));
+	set((zlong)((a && !b) || (!a && b)));
 	break;
     case COMMA:
 	pop2();
@@ -725,18 +727,18 @@ bop(int tk)
 
 
 /**/
-static long
+static zlong
 mathevall(char *s, int prek, char **ep)
 {
     int t0;
     int xlastbase, xnoeval, xunary, xlvc;
     char *xptr;
-    long xyyval;
+    zlong xyyval;
     LV xyylval;
-    char **xlvals = 0;
+    char **xlvals = 0, *nlvals[LVCOUNT];
     int xsp;
-    struct mathvalue *xstack = 0;
-    long ret;
+    struct mathvalue *xstack = 0, nstack[STACKSZ];
+    zlong ret;
 
     xlastbase = xnoeval = xunary = xlvc = xyyval = xyylval = xsp = 0;
     xptr = NULL;
@@ -753,9 +755,10 @@ mathevall(char *s, int prek, char **ep)
 	xsp = sp;
 	xstack = stack;
     }
-    stack = (struct mathvalue *)zalloc(STACKSZ*sizeof(struct mathvalue));
+    stack = nstack;
     lastbase = -1;
-    lvals = (char **)zcalloc(LVCOUNT*sizeof(char *));
+    memset(nlvals, 0, LVCOUNT*sizeof(char *));
+    lvals = nlvals;
     lvc = 0;
     ptr = s;
     sp = -1;
@@ -769,8 +772,6 @@ mathevall(char *s, int prek, char **ep)
 
     ret = stack[0].val;
 
-    zfree(lvals, LVCOUNT*sizeof(char *));
-    zfree(stack, STACKSZ*sizeof(struct mathvalue));
     if (--mlevel) {
 	lastbase = xlastbase;
 	noeval = xnoeval;
@@ -789,11 +790,11 @@ mathevall(char *s, int prek, char **ep)
 
 
 /**/
-long
+zlong
 matheval(char *s)
 {
     char *junk;
-    long x;
+    zlong x;
     int xmtok = mtok;
 
     if (!*s)
@@ -807,10 +808,10 @@ matheval(char *s)
 
 
 /**/
-long
+zlong
 mathevalarg(char *s, char **ss)
 {
-    long x;
+    zlong x;
     int xmtok = mtok;
 
     x = mathevall(s, ARGPREC, ss);
@@ -827,7 +828,8 @@ mathevalarg(char *s, char **ss)
 static void
 mathparse(int pc)
 {
-    int q, otok, onoeval;
+    zlong q;
+    int otok, onoeval;
 
     if (errflag)
 	return;
